/*
Copyright (c) [2023] [Ben Hepple]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

///////////////////////////
// Important Information //
///////////////////////////

/*

NOTE: This is a LEGACY Code build that is no longer used by Coalescence Greenpower.
      It is not well commented and I recommend looking at Ajax 2 V4 hardware and 
      software as a better example. I have only provided this as an example to 
      work with the Ajax 2 V2 hardware.

      This code example is an old 'Race Certified Program' that has had some code 
      removed for safety. Other than that this is real code that propelled us to 
      a 2nd place result at Lotus Hethel in 2023. 

      In this 'Base Build' you will find the bare minimum needed for the Ajax 2 V2
      board to run correctly. I have included the 'Start' Mode Functions to save 
      you the time of working out the control logic required to have soft starts.

      If you wish to add code and don't know where to begin or you don't understand 
      why I have done something a certain way please do not hesitate to contact the 
      team and I at: coalescence.gp@gmail.com or www.coalescencegp.co.uk

We are very friendly and always willing to help. 

Ben 

*/

/////////////////////////////////////////////////////

/*           PWM and Throttle Setup                */

/////////////////////////////////////////////////////

float startLimit        = 38;
float currentLimit1     = 31;
float currentLimit2     = 33;
float currentLimit3     = 35;
float bypassLimit       = 38;

float Kp = 0.35;

#define pwmPin    27
#define throttle1 17

int       throttle    = 0;
const int freq        = 15000;
const int resolution  = 8;
const int pwmChannel  = 0;
int       pwm         = 0;

int       carMode     = 1;
int       driveMode   = 1;

const int carModeInterval   = 400;
unsigned long carModeMillis = 0;

bool startButtonTimer = true;
const int waitKeyInt  = 4000; 
unsigned long waitKey = 0;

bool firstTouch       = true;
bool rampInc          = false;
bool TFlag            = true;
int  TPos             = 0;  


int totalError = 0;
int previousError = 0;
float u;

/////////////////////////////////////////////////////

/*       SD Setup variables and Functions          */

/////////////////////////////////////////////////////

#include "FS.h"
#include "SD.h"
#include "SPI.h"

unsigned long dataMillis = 0;
const int   dataInterval = 100;
unsigned long raceTime   = 0;
String dataString; 

  // SD Card command functions
  
void appendTitle(fs::FS &fs, const char * path, const char * message)
{
  File file = fs.open(path, FILE_APPEND);
  file.print(message);
  file.close();
}

void appendData(fs::FS &fs, const char * path, String  message)
{
  File file = fs.open(path, FILE_APPEND);
  file.print(message);
  file.close();
}

/////////////////////////////////////////////////////

/*           Current and Voltage Variables         */

/////////////////////////////////////////////////////

#define V_Batt  26
#define V_Motor 13
#define Amp_In  33

float batteryVoltage        = 0;              
float motorVoltage          = 0;
float Current               = 0;
float convert               =0;

const int offsetSmoothing  = 40;
const int currentSmoothing = 50;
int       currentOffset    = 0;
float     mapb             = 34.2;    //milli amps per bit        

/////////////////////////////////////////////////////

/*           Defining Display Buttons              */

/////////////////////////////////////////////////////

#define driverIn1 4
#define driverIn2 16

int input1 = 0;
int input2 = 0;



uint8_t displayPause = 0;

/////////////////////////////////////////////////////

/*           Defining RPM Interrupt                */

/////////////////////////////////////////////////////

int RPM = 0;
const uint8_t rpmPin = 25;
volatile unsigned long motorPoll = 0;
unsigned long lastMotorSpeedPollTime = 0;

void IRAM_ATTR isr() {
  motorPoll++;
}

/////////////////////////////////////////////////////


/*                      SETUP                      */


/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

void setup() 
{
  Serial.begin(115200);                                                         // Any baud rate lower than 115200 is inadequate 
  
  // Setup Driver Inputs
  pinMode(throttle1, INPUT);
  pinMode(driverIn1, INPUT);
  pinMode(driverIn2, INPUT);
  //pinMode(input3, INPUT);
  //pinMode(input4, INPUT);
  
  // Setup PWM Channel   
  ledcSetup(pwmChannel, freq, resolution);
  ledcAttachPin(pwmPin, pwmChannel);                                            // Attach the LED PWM Channel to the GPIO Pin 
  ledcWrite(pwmChannel, 0);                                                     // Set initial pwm value to 0 so no drive is active

  // Current Sensor Calibration 
  int adcAverage = 0;
  for (int i = 0; i<offsetSmoothing; i++){ 
    adcAverage = adcAverage + analogRead(Amp_In);
  }
   currentOffset = adcAverage/offsetSmoothing;

  // RPM Interrupt
  attachInterrupt(rpmPin, isr, LOW);

  // Setup SD card 
  if(!SD.begin(5)){
    Serial.println("Card Mount Failed");
  }
  appendTitle(SD, "/HadesData.txt", "TIME, BV, MV, C, RPM");
  


  // initiate soft start checking
  waitKey = millis();
}

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////


/*                  Main Loop                      */


/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

void loop() 
{ 

  input1 = digitalRead(driverIn1);
  input2 = digitalRead(driverIn2);
  readCurrent();
  readBatteryVoltage();
  readMotorVoltage();
  timer();

  if((millis() > (waitKey + waitKeyInt)) && Current <= 2 && motorVoltage <= 3.5){
    driveMode = 0;
  } 
  else if(motorVoltage >= 19){
    driveMode = carMode;
  }


  throttle = digitalRead(throttle1);

  if(carMode == 4){
    driveMode = carMode;
  }

  if(throttle == LOW)
  {
    if(TFlag == false){
      TPos = 0;
      TFlag = true; 
    }
    
    startButtonTimer = true;
    switch(driveMode){
      
      case 0: 
        startMode();
        break;

      case 1:
        pwm += PIDu(currentLimit1);
        break;

      case 2:
        pwm += PIDu(currentLimit2);
        break;

      case 3:
        pwm += PIDu(currentLimit3);
        break; 
        
      case 4:
        pwm += PIDu(bypassLimit);
        break;
    }
  }
  if(throttle == HIGH)
  {
    pwm = 0;
      
    if (TFlag == true){
      TPos = 1;
    }
    else if(TFlag == false){
      TPos = 0;
    }
     
    
    if(startButtonTimer == true){
      startButtonTimer = false;
      waitKey = millis();    
    }   
  }
  if(pwm>=255){
    pwm = 255;
  }
  else if(pwm < 0){
    pwm = 0;
  }
 ledcWrite(pwmChannel, pwm);
  
}

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////


/*                 Functions                       */             // These are the blocks of code that allow data to be gathered, read and written for all the boards functions
                                                                  

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

void startMode(){ 
  if(firstTouch == true){
    firstTouch = false; 
    pwm = 130;
  }
  
  if(rampInc == true && Current <= startLimit){
    rampInc = false;
    pwm = pwm + 5;
  }
  
}

int PIDu(int limit){
  float error = limit - Current;
  float u1;
  u1 = Kp * error; 
  return u1;
}

void readCurrent (){                                        
  int adcAverage = 0;
  for (int i = 0; i<currentSmoothing; i++)
  { 
    adcAverage = adcAverage + analogRead(Amp_In);
  }
  float tempCurrent = adcAverage/currentSmoothing;
  //Serial.print(tempCurrent);
  tempCurrent = tempCurrent - currentOffset;
  tempCurrent = constrain(tempCurrent,0,4096);
  //Serial.print("    Offset: "); Serial.println(tempCurrent);
  tempCurrent = tempCurrent * mapb;
  
  Current = tempCurrent/1000;
}


void readBatteryVoltage()
{ 
   float tempVoltage = analogRead(V_Batt);                        // Reads the 10 Bit ADC Value from the battery read pin (A0)

                                                                  // Calculates the voltage seen on the A0 Pin (Up to 5V given we are using the TSR 1-2450)
   tempVoltage = (tempVoltage/4096)*3.3;                          // This is a multiplier which is related to the potential divider ratio. 
   tempVoltage = (tempVoltage * 9.09);
                                                                  // For us, this ratio is 1/6 which means we step the voltage from 30V --> 5V so the arduino can read. 
                                                                                                                                   
   tempVoltage = constrain(tempVoltage,0,30);                     // Gets rid of zero errors so the LCD display doesn't bug out
   batteryVoltage = tempVoltage;                                  // Returns the voltage to the function readBatteryVoltage() 
   //Serial.println(batteryVoltage); 
}



void readMotorVoltage()
{
  
   float tempVoltage = analogRead(V_Motor);                       // Reads the 10 Bit ADC Value from the battery read pin (A0)

                                                                  // Calculates the voltage seen on the A0 Pin (Up to 5V given we are using the TSR 1-2450)
   tempVoltage = (tempVoltage/4096)*3.3;                          // This is a multiplier which is related to the potential divider ratio. 
   tempVoltage = (tempVoltage * 9.06);
                                                                  // For us, this ratio is 1/6 which means we step the voltage from 30V --> 5V so the arduino can read. 
                                                                  // Gets rid of zero errors so the LCD display doesn't bug out
   tempVoltage = batteryVoltage - tempVoltage;
   float value = constrain(tempVoltage,0,30);
   motorVoltage = value;                                          // Returns the voltage to the function readBatteryVoltage()
   //Serial.println(motorVoltage);
}

void readMotorRPM()
{
 
  int tempMotorPoll = motorPoll;                                                    // Records the number of interrupts that have occured since the last function recall  
  motorPoll = 0;                                                                    // Reset previous value
  
  unsigned long tempLastMotorPollTime = lastMotorSpeedPollTime;                     // Use variables to remember the time when the last RPM was calculated.
  
  lastMotorSpeedPollTime = millis();                                                // Record the current time to be used in the this cycle and the next.
  float motorRevolutions = tempMotorPoll / 4;                                       // Now calculate the number of revolutions of the motor shaft
  
  motorRevolutions = motorRevolutions * 60.0;                                       // Calculate the revolutions in one minute
  float timeDifference = (lastMotorSpeedPollTime - tempLastMotorPollTime)/1000.0;   // Find the time difference in seconds
    
  float ShaftRPM = motorRevolutions / timeDifference;                               // Finally calculate RPM 
  RPM = ShaftRPM;
  
}

void timer()                                                                  // Seperate function for timers
{
  
  if (millis() >= carModeMillis + carModeInterval)  
  {
    carModeMillis += carModeInterval;
    rampInc = true;
    int tempMode = carMode;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               
    if((input1 == HIGH) && (input2 == HIGH)){
      carMode = 1;
    }
    else if((input1 == HIGH) && (input2 == LOW)){
      carMode = 2;
    }
    else if((input1 == LOW) && (input2 == HIGH)){
      carMode = 3;
    }
    else if((input1 == LOW) && (input2 == LOW)){
      carMode = 4;
    }
  
  }

  if (millis() >= dataMillis + dataInterval)  
  {
    dataMillis += dataInterval;
    saveData(); 
  }

  if(motorPoll >= 28){
    readMotorRPM();
  }
}

void saveData()
{
  raceTime = millis();

  String (dataString) = "\n" + String(raceTime) + "," + String(batteryVoltage) + "," + String(motorVoltage) + "," + String(Current) + "," + String(RPM);
  appendData(SD, "/HadesData.txt", dataString);
  
}
