/////////////////////////////////////////////////////

/*           Discharge Characteristics             */

/////////////////////////////////////////////////////

// user variables
const float dischargeLimit = 10.7;                    // single battery votlage limit
const int   numOfBatteries = 2;                       // Number of 12V batteries in series
const int dischargeRate    = 170;                    // 0 - 255 (increasing value = increase rate)
const int maxCurrent       = 15; 

int rampRate = 2;                                     // value between 1 and 5

// private variables
int startRamp             = 1; 
int ramp                  = 50;
unsigned long  rampMillis = 0;
const int    rampInterval = 250;
float     calculatedLimit = 0;                           


/////////////////////////////////////////////////////

/*           PWM and Throttle Setup                */

/////////////////////////////////////////////////////

#define pwmPin    27
const int freq        = 15000;
const int resolution  = 8;
const int pwmChannel  = 0;
int       pwm         = 0;

/////////////////////////////////////////////////////

/*       SD Setup variables and Functions          */

/////////////////////////////////////////////////////

#include "FS.h"
#include "SD.h"
#include "SPI.h"

unsigned long dataMillis = 0;
const int   dataInterval = 1000;
unsigned long raceTime   = 0;

unsigned long serialMillis = 0;
const int    serialInterval = 2000;

String dataString; 

  // SD Card command functions
  
void appendTitle(fs::FS &fs, const char * path, const char * message)
{
  File file = fs.open(path, FILE_APPEND);
  file.print(message);
  file.close();
}

void appendData(fs::FS &fs, const char * path, String  message)
{
  File file = fs.open(path, FILE_APPEND);
  file.print(message);
  file.close();
}

/////////////////////////////////////////////////////

/*           Current and Voltage Variables         */

/////////////////////////////////////////////////////

#define V_Batt  26
#define V_Motor 13
#define Amp_In  33


float   ampHours = 0;
long  pasttime = 0;      // stores time since last update to ampHours

const int offsetSmoothing  = 40;
const int currentSmoothing = 50;
int       currentOffset    = 0;
float     mapb             = 34.2;    //milli amps per bit        


/////////////////////////////////////////////////////

/////////////////////////////////////////////////////


/*                      SETUP                      */


/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

void setup() 
{
  Serial.begin(115200);                                                         // Any baud rate lower than 115200 is inadequate 

  // Setup battery characterstics
  calculatedLimit = dischargeLimit*numOfBatteries;
  
  // Setup PWM Channel   
  ledcSetup(pwmChannel, freq, resolution);
  ledcAttachPin(pwmPin, pwmChannel);                                            // Attach the LED PWM Channel to the GPIO Pin 
  ledcWrite(pwmChannel, 0);                                                     // Set initial pwm value to 0 so no drive is active

  // Current Sensor Calibration 
  int adcAverage = 0;
  for (int i = 0; i<offsetSmoothing; i++){ 
    adcAverage = adcAverage + analogRead(Amp_In);
  }
   currentOffset = adcAverage/offsetSmoothing;

  // Setup SD card 
  if(!SD.begin(5)){
    Serial.println("Card Mount Failed");
    //while(1);
  }
  appendTitle(SD, "/HadesData.txt", "TIME, BV, MV, C, pwm, AH");

  delay(6000);

  if(startRamp == 1){
    Serial.println("---------------------------");
    Serial.println("SYSTEM: RAMP INITIATED     ");
    Serial.println("Discharge Limit: ");
    Serial.println(calculatedLimit);
    Serial.println("---------------------------");
  }
}

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////


/*                  Main Loop                      */


/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

void loop() 
{ 

  float Current = readCurrent();
  float batteryVoltage = readBatteryVoltage();
  float motorVoltage = readMotorVoltage();
  float SOC = readAmpHours();
  timer();

  
  

  if(startRamp == 1){
      if (millis() >= rampMillis + rampInterval){
        rampMillis += rampInterval;
        ramp += rampRate;
        pwm = ramp;
      }
  }
  else{
    pwm = dischargeRate;
  }
  
  if(startRamp == 1 && ramp >= dischargeRate){
    startRamp = 0;
    pwm = dischargeRate;
    Serial.println("SYSTEM: RAMP COMPLETE      ");
    Serial.print("SYSTEM: HOLDING AT -  ");
    Serial.println(dischargeRate);
    Serial.println("---------------------------");
  }

  if(Current >= maxCurrent){
    pwm = 0; 
    Serial.println("ERROR: CURRENT MAX EXCEEDED");
    Serial.println("---------------------------");
  }
  /*
  if(batteryVoltage < calculatedLimit){
    pwm = 0; 
    Serial.println("SYSTEM: DISCHARGE COMPLETE");
    Serial.println("---------------------------");
    saveData(Current,batteryVoltage,motorVoltage,SOC);
    updateSerial(Current,batteryVoltage,motorVoltage,SOC);
    ledcWrite(pwmChannel, pwm);
    while(1); 
  }
*/
  ledcWrite(pwmChannel, pwm);
  
}

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////


/*                 Functions                       */             // These are the blocks of code that allow data to be gathered, read and written for all the boards functions
                                                                  

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

float readCurrent (){                                        
  int adcAverage = 0;
  for (int i = 0; i<currentSmoothing; i++)
  { 
    adcAverage = adcAverage + analogRead(Amp_In);
  }
  float tempCurrent = adcAverage/currentSmoothing;
  //Serial.print(tempCurrent);
  tempCurrent = tempCurrent - currentOffset;
  tempCurrent = constrain(tempCurrent,0,4096);
  //Serial.print("    Offset: "); Serial.println(tempCurrent);
  tempCurrent = tempCurrent * mapb;
  
  tempCurrent = tempCurrent/1000;
  return(tempCurrent);
  //Serial.println(Current);
}


float readAmpHours()
{ 
  float Current = readCurrent();
  long nowtime = millis();
  float difference = nowtime - pasttime;
  pasttime = nowtime;
  
  float Atime = (Current*1000) * difference;
  
  float Ahours = (Atime/1000)/3600;
  
  ampHours = ampHours + Ahours;
  float returnAH = ampHours/1000;
  return (returnAH);
}


float readBatteryVoltage()
{ 
   
  int voltageSmoothing = 50;
  int adcTotal = 0;
  for (int i = 0; i<voltageSmoothing; i++)
  { 
    adcTotal = adcTotal + analogRead(V_Batt);
  }
  float tempVoltage = adcTotal/voltageSmoothing;
                                                                  // Calculates the voltage seen on the bv Pin
   tempVoltage = (tempVoltage/4096)*3.3;                          // This is a multiplier which is related to the potential divider ratio. 
   tempVoltage = (tempVoltage * 9.09);
   tempVoltage = (tempVoltage + 1.3);
                                                                  // For us, this ratio is 1/9 which means we step the voltage from 30V --> 3.3V so the arduino can read. 
                                                                                                                                   
   tempVoltage = constrain(tempVoltage,0,30);                     // Gets rid of zero errors so the display doesn't bug out
   return (tempVoltage);                                          // Returns the voltage to the function readBatteryVoltage() 
   //Serial.println(batteryVoltage); 
}



float readMotorVoltage()
{
   float batteryVoltage = readBatteryVoltage();
   float tempVoltage = analogRead(V_Motor);                       // Reads the 12 Bit ADC Value from the battery read pin (A0)

                                                                  
   tempVoltage = (tempVoltage/4096)*3.3;                          // Calculates the voltage seen on the mv Pin                   
   tempVoltage = (tempVoltage * 9.09);                            // This is a multiplier which is related to the potential divider ratio.
   tempVoltage = tempVoltage + 0.6;                                
                                                                  
   tempVoltage = batteryVoltage - tempVoltage;
   float value = constrain(tempVoltage,0,30);
   return (value);                                          // Returns the voltage to the function readBatteryVoltage()
   //Serial.println(motorVoltage);
}





void timer()                                                                  // Seperate function for timers
{
  if (millis() >= serialMillis + serialInterval)
  {
    serialMillis += serialInterval;
    float Current = readCurrent();
    float batteryVoltage = readBatteryVoltage();
    float motorVoltage = readMotorVoltage();
    float SOC = readAmpHours();
    updateSerial(Current,batteryVoltage,motorVoltage,SOC);
  }

  if (millis() >= dataMillis + dataInterval)  
  {
    dataMillis += dataInterval;
    float Current = readCurrent();
    float batteryVoltage = readBatteryVoltage();
    float motorVoltage = readMotorVoltage();
    float SOC = readAmpHours();
    saveData(Current,batteryVoltage,motorVoltage,SOC);
  }
}

void updateSerial(float Current,float batteryVoltage,float motorVoltage,float SOC)
{
  float watts = Current*motorVoltage;
  Serial.print("          OV: ");
  Serial.println(motorVoltage);
  Serial.print("          BV: ");
  Serial.println(batteryVoltage);
  Serial.print("           C: ");
  Serial.println(Current);
  Serial.print("          Ah: ");
  Serial.println(SOC);
  Serial.print("       WATTS: ");
  Serial.println(watts);
  Serial.println("---------------------------");
}

void saveData(float Current,float batteryVoltage,float motorVoltage,float SOC)
{
  raceTime = millis();

  String (dataString) = "\n" + String(raceTime) + "," + String(batteryVoltage) + "," + String(motorVoltage) + "," + String(Current) + "," + String(pwm) + "," + String(SOC);
  appendData(SD, "/HadesData.txt", dataString);

}
