#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Fonts/FreeMonoBoldOblique18pt7b.h>
#include <Adafruit_SSD1306.h>
/////////////////////////////////////////////////////

/*                Custom variables                 */

/////////////////////////////////////////////////////

uint8_t currentLimit = 28;
uint8_t debugPage    = 0;

bool boostToggle = true;

/////////////////////////////////////////////////////

/*           Display Setup variables               */

/////////////////////////////////////////////////////

#define SCREEN_WIDTH 128                                                       // OLED display width, in pixels
#define SCREEN_HEIGHT 64                                                       // OLED display height, in pixels
#define OLED_RESET     -1                                                      // Reset pin # (or -1 if sharing Arduino reset pin)
#define SCREEN_ADDRESS 0x3C                                                    // See back of display for resistor setting (either 0x3C or 0x3D)

int Current;


Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

/////////////////////////////////////////////////////

/*           Defining Button Interrupts            */

/////////////////////////////////////////////////////

#define throttle1 17
#define throttle2 18

uint8_t dispSetting = 0
uint8_t updateVal   = 0;
uint8_t screenNum   = 3;

struct Button {                                                                 //Struct Button is a function that creates a list of variables under a single name  
  const uint8_t pin;                                                            //This makes it easier to setup multiple variable values on one command as shown below               
  bool pressed;
};

Button disp1 = {4, false};                                                      // Define disp1 onto GPIO 4   (Calling Button before a variable name now allows values to be assigned to pin and pressed(state))
Button disp2 = {0, false};                                                      // Define disp2 onto GPIO 0   
Button disp3 = {12,false};                                                      // Define disp3 onto GPIO 8

void IRAM_ATTR isr1()                                                           // Attach 3 seperate interrupts so that each button has a steady state bool variable that stops debouncing 
{
  disp1.pressed = true;                                                         // Have a bool toggle true/false (See dispChange() function) allows all switch bounces to be eliminated 
}                                                                               // however, this relies upon the void loop() function running with 0 delays
void IRAM_ATTR isr2()
{
  disp2.pressed = true;
}
void IRAM_ATTR isr3()
{
  disp3.pressed = true;
}

void setup() 
{
  Serial.begin(9600);

  
  if(!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS))                      // SSD1306_SWITCHCAPVCC = generate display voltage from 3.3V internally
  {                    
    Serial.println(F("SSD1306 allocation failed"));
    //for(;;);                                                                  // Don't proceed, loop forever (Uncomment if you want program to crash lol)
  }
  display.setFont(&FreeMonoBoldOblique18pt7b);                                  // IMPORTANT, This command sets the font and size of text shown on display. Please read Adafruit GFX documentation.
  display.display();
  
  delay(1000);                                                                  // Pause for 1 seconds

  display.clearDisplay();
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0,45);
  display.print("AJAX 2");
  display.display();

  delay(1000);

  pinMode(throttle1, INPUT);
  pinMode(throttle2, INPUT);
  pinMode(disp1.pin, INPUT);
  pinMode(disp2.pin, INPUT);
  pinMode(disp3.pin, INPUT);

  attachInterrupt(disp1.pin, isr1, LOW);
  attachInterrupt(disp2.pin, isr2, LOW);
  attachInterrupt(disp3.pin, isr3, LOW);


}

void loop() 
{
  
  dispUpdate();
  Serial.println(dispSetting);


}

void dispUpdate()
{

  updateVal = 0;

  if(disp1.pressed == true)
  {
    disp1.pressed =false;
    updateVal = 1  
  }
  else if(disp2.pressed == true)
  {
    disp2.pressed = false;
    updateVal = 2  
  }
  else if(disp3.pressed == true)
  {
    disp3.pressed = false;
    dispSetting += 1
    if(dispSetting > screenNum)
    {
      dispSetting = 1;
    }
  }



  switch (dispSetting)
  {
    case 1:    
      if(updateVal == 1)
      {
        currentLimit -= 1;
      }  
      else if(updateVal == 2)
      {
        currentLimit += 1; 
      }
      screen1();
      break;

    case 2:
      if(updateVal == 1 || updateVal == 2)
      {
         if(boostToggle == true){
           boostToggle = false;
         }
         else{
           boostToggle = true;
         }
      }
      screen2();
      break;

    case 3:     
      screen3();
      break;
  }

}

void screen1(void) 
{
  display.clearDisplay(); 
  display.setCursor(0,40);                                                                                  
  display.setTextColor(SSD1306_WHITE);
  display.print(F("L:")); display.print(currentLimit);
  display.display();  
}

void screen2(void)
{
  display.clearDisplay(); 
  display.setCursor(0,40);                                                                                  
  display.setTextColor(SSD1306_WHITE);
  display.print(F("A:")); display.print(Current);
  display.display();
}

void screen3(void)
{
  display.clearDisplay(); 
  display.setCursor(0,40);                                                                                  
  display.setTextColor(SSD1306_WHITE);
  display.print(F("A:")); display.print(Current);
  display.display();
  delay(200);
}


